"""Contains the extension setup.

"""

from pathlib import Path
from typing import Dict, Any

from sphinx.domains import ObjType
from sphinx.application import Sphinx

from sphinxcontrib.autodoc_pydantic.autodoc import (
    PydanticValidatorDocumenter,
    PydanticModelDocumenter,
    PydanticConfigClassDocumenter,
    PydanticFieldDocumenter,
    PydanticSettingsDocumenter
)

from sphinxcontrib.autodoc_pydantic.directives import (
    PydanticField,
    PydanticConfigClass,
    PydanticValidator,
    PydanticModel,
    PydanticSettings
)

__version__ = "1.1.3"


def add_css_file(app: Sphinx, exception: Exception):
    """Adds custom css to HTML output.

    """

    filename = "autodoc_pydantic.css"
    static_path = (Path(app.outdir) / "_static").absolute()
    static_path.mkdir(exist_ok=True, parents=True)
    path_css = Path(__file__).parent.joinpath("css", filename)

    if not (static_path / filename).exists():
        content = path_css.read_text()
        (static_path / filename).write_text(content)


def add_domain_object_types(app: Sphinx):
    """Hack to add object types to already instantiated python domain since
    `add_object_type` currently only works for std domain.

    """

    object_types = app.registry.domain_object_types.setdefault("py", {})

    obj_types = ["pydantic_model",
                 "pydantic_settings",
                 "pydantic_field",
                 "pydantic_validator",
                 "pydantic_config"]

    for obj_type in obj_types:
        object_types[obj_type] = ObjType(obj_type, "obj", "any")


def add_configuration_values(app: Sphinx):
    """Adds all configuration values to sphinx application.

    """

    stem = "autodoc_pydantic_"
    add = app.add_config_value

    add(f'{stem}config_signature_prefix', "model", True, str)
    add(f'{stem}config_members', True, True, bool)

    add(f'{stem}settings_show_json', True, True, bool)
    add(f'{stem}settings_show_config_member', False, True, bool)
    add(f'{stem}settings_show_config_summary', True, True, bool)
    add(f'{stem}settings_show_validator_members', True, True, bool)
    add(f'{stem}settings_show_validator_summary', True, True, bool)
    add(f'{stem}settings_hide_paramlist', True, True, bool)
    add(f'{stem}settings_undoc_members', True, True, bool)
    add(f'{stem}settings_members', True, True, bool)
    add(f'{stem}settings_member_order', 'groupwise', True, str)
    add(f'{stem}settings_signature_prefix', "pydantic settings", True, str)

    add(f'{stem}model_show_json', True, True, bool)
    add(f'{stem}model_show_config_member', False, True, bool)
    add(f'{stem}model_show_config_summary', True, True, bool)
    add(f'{stem}model_show_validator_members', True, True, bool)
    add(f'{stem}model_show_validator_summary', True, True, bool)
    add(f'{stem}model_hide_paramlist', True, True, bool)
    add(f'{stem}model_undoc_members', True, True, bool)
    add(f'{stem}model_members', True, True, bool)
    add(f'{stem}model_member_order', 'groupwise', True, str)
    add(f'{stem}model_signature_prefix', "pydantic model", True, str)

    add(f'{stem}validator_signature_prefix', "validator", True, str)
    add(f'{stem}validator_replace_signature', True, True, bool)
    add(f'{stem}validator_list_fields', False, True, bool)

    add(f'{stem}field_list_validators', True, True, bool)
    add(f'{stem}field_doc_policy', 'both', True, str)
    add(f'{stem}field_show_constraints', True, True, bool)
    add(f'{stem}field_show_alias', True, True, bool)
    add(f'{stem}field_show_default', True, True, bool)
    add(f'{stem}field_signature_prefix', "field", True, str)


def add_directives_and_autodocumenters(app: Sphinx):
    """Adds custom pydantic directives and autodocumenters to sphinx
    application.

    """

    app.add_directive_to_domain("py", "pydantic_field", PydanticField)
    app.add_directive_to_domain("py", "pydantic_model", PydanticModel)
    app.add_directive_to_domain("py", "pydantic_settings", PydanticSettings)
    app.add_directive_to_domain("py", "pydantic_config", PydanticConfigClass)
    app.add_directive_to_domain("py", "pydantic_validator", PydanticValidator)

    app.setup_extension('sphinx.ext.autodoc')  # Require autodoc extension
    app.add_autodocumenter(PydanticFieldDocumenter)
    app.add_autodocumenter(PydanticModelDocumenter)
    app.add_autodocumenter(PydanticSettingsDocumenter)
    app.add_autodocumenter(PydanticValidatorDocumenter)
    app.add_autodocumenter(PydanticConfigClassDocumenter)


def setup(app: Sphinx) -> Dict[str, Any]:
    add_configuration_values(app)
    add_directives_and_autodocumenters(app)
    add_domain_object_types(app)
    app.add_css_file("autodoc_pydantic.css")
    app.connect("build-finished", add_css_file)

    return {
        'version': __version__,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }


"""This module contains directive utilities like option validation functions
and the PyAutoDoc composite class.

"""
import functools
from typing import Any, Union, List, Set, Callable, Optional

from docutils.nodes import emphasis
from docutils.parsers.rst import Directive
from sphinx.addnodes import pending_xref
from sphinx.environment import BuildEnvironment
from sphinx.ext.autodoc import ALL, Documenter

from sphinxcontrib.autodoc_pydantic.inspection import NamedReference


class NullType:
    """Helper class to present a Null value which is not the same
    as python's `None`. This represents a missing value, or no
    value at all by convention. It should be used as a singleton.

    """

    def __bool__(self):
        return False

NONE = NullType()


def create_field_href(reference: NamedReference,
                      env: BuildEnvironment) -> pending_xref:
    """Create `pending_xref` node with link to given `reference`.

    """

    text = reference.name
    options = {'refdoc': env.docname,
               'refdomain': "py",
               'reftype': "obj",
               'reftarget': reference.ref}

    refnode = pending_xref(reference.name, **options)
    classes = ['xref', "py", '%s-%s' % ("py", "obj")]
    refnode += emphasis(text, text, classes=classes)
    return refnode


def remove_node_by_tagname(nodes: List, tagname: str):
    """Removes node from list of `nodes` with given `tagname` in place.

    """

    for remove in [node for node in nodes if node.tagname == tagname]:
        nodes.remove(remove)


def option_members(arg: Any) -> Union[object, List[str]]:
    """Used to convert the :members: option to auto directives.

    """

    if isinstance(arg, str):
        sanitized = arg.lower()
        if sanitized == "true":
            return ALL
        elif sanitized == "false":
            return None

    if arg in (None, True):
        return ALL
    elif arg is False:
        return None
    else:
        return [x.strip() for x in arg.split(',') if x.strip()]


def option_one_of_factory(choices: Set[Any]) -> Callable:
    """Create a option validation function which allows only one value
    of given set of provided `choices`.

    """

    def option_func(value: Any):
        if value not in choices:
            raise ValueError(f"Option value has to be on of {choices}")
        return value

    return option_func


def option_default_true(arg: Any) -> bool:
    """Used to define boolean options with default to True if no argument
    is passed.

    """

    if isinstance(arg, bool):
        return arg

    if arg is None:
        return True

    sanitized = arg.strip().lower()

    if sanitized == "true":
        return True
    elif sanitized == "false":
        return False
    else:
        raise ValueError(f"Directive option argument '{arg}' is not valid. "
                         f"Valid arguments are 'true' or 'false'.")


def option_list_like(arg: Any) -> Set[str]:
    """Used to define a set of items.

    """

    if not arg:
        return set()
    else:
        return {x.strip() for x in arg.split(",")}


class PydanticAutoDirective:
    """Composite class providing methods to handle getting and setting
    directive option values.

    """

    def __init__(self, parent: Union[Documenter, Directive]):
        self.parent = parent
        self.add_default_options()

    def add_default_options(self):
        """Adds all default options.

        """

        options = getattr(self.parent, "pyautodoc_set_default_option", [])
        for option in options:
            self.set_default_option(option)

    @staticmethod
    def sanitize_configuration_option_name(name: str) -> str:
        """Provide full app environment configuration name for given option
        name while converting "-" to "_".

        Parameters
        ----------
        name: str
            Name of the option.

        Returns
        -------
        full_name: str
            Full app environment configuration name.

        """

        sanitized = name.replace("-", "_")

        return f"autodoc_pydantic_{sanitized}"

    def is_available(self, name: str) -> bool:
        """Check if option is usable.

        """

        available = self.parent.options.get("__doc_disable_except__")
        if available is None:
            return True
        else:
            return name in available

    def get_configuration_by_name(self, name: str) -> Any:
        """Get configuration value from app environment configuration.
        If `name` does not exist, return NONE.

        """

        config_name = self.sanitize_configuration_option_name(name)
        return getattr(self.parent.env.config, config_name, NONE)

    def get_option_value(self, name: str, prefix: bool = False) -> Any:
        """Get option value for given `name`. First, looks for explicit
        directive option values (e.g. :member-order:) which have highest
        priority. Second, if no directive option is given, get the default
        option value provided via the app environment configuration.

        Parameters
        ----------
        name: str
            Name of the option.
        prefix: bool
            If True, add `pyautodoc_prefix` to name.

        """

        if prefix:
            name = f"{self.parent.pyautodoc_prefix}-{name}"

        if name in self.parent.options:
            return self.parent.options[name]
        elif self.is_available(name):
            return self.get_configuration_by_name(name)

    def option_is_false(self, name: str, prefix: bool = False) -> bool:
        """Get option value for given `name`. First, looks for explicit
        directive option values (e.g. :member-order:) which have highest
        priority. Second, if no directive option is given, get the default
        option value provided via the app environment configuration.

        Enforces result to be either True or False.

        Parameters
        ----------
        name: str
            Name of the option.
        prefix: bool
            If True, add `pyautodoc_prefix` to name.

        """

        return self.get_option_value(name=name, prefix=prefix) is False

    def option_is_true(self, name: str, prefix: bool = False) -> bool:
        """Get option value for given `name`. First, looks for explicit
        directive option values (e.g. :member-order:) which have highest
        priority. Second, if no directive option is given, get the default
        option value provided via the app environment configuration.

        Enforces result to be either True or False.

        Parameters
        ----------
        name: str
            Name of the option.
        prefix: bool
            If True, add `pyautodoc_prefix` to name.

        """

        return self.get_option_value(name=name, prefix=prefix) is True


    def set_default_option(self, name: str):
        """Set default option value for given `name` from app environment
        configuration if an explicit directive option was not provided.

        Parameters
        ----------
        name: str
            Name of the option.

        """

        if (name not in self.parent.options) and (self.is_available(name)):
            self.parent.options[name] = self.get_configuration_by_name(name)

    def set_default_option_with_value(self, name: str,
                                      value_true: Any,
                                      value_false: Optional[Any] = NONE):
        """Set option value for given `name`. Depending on app environment
        configuration boolean value choose either `value_true` or
        `value_false`. This is only relevant if option value has not been set,
        yet.

        Parameters
        ----------
        name: str
            Name of the option.
        value_true:
            Value to be set if True.
        value_false:
            Value to be set if False.

        """

        value = self.parent.options.get(name)

        if value is not None and self.is_available(name):
            if self.get_configuration_by_name(name):
                self.parent.options[name] = value_true
            elif value_false is not NONE:
                self.parent.options[name] = value_false


class PydanticAutoDoc(PydanticAutoDirective):
    """Composite class providing methods to handle getting and setting
    autodoc directive option values.

    """

    def __init__(self, *args):
        super().__init__(*args)
        self.add_pass_through_to_directive()

    def sanitize_configuration_option_name(self, name: str) -> str:
        """Provide full app environment configuration name for given option
        name.

        Parameters
        ----------
        name: str
            Name of the option.

        Returns
        -------
        full_name: str
            Full app environment configuration name.

        """

        sanitized = name.replace("-", "_")
        prefix = self.parent.objtype.split("_")[-1]

        if prefix not in sanitized:
            sanitized = f"{prefix}_{sanitized}"

        return f"autodoc_pydantic_{sanitized}"

    def add_pass_through_to_directive(self):
        """Intercepts documenters `add_directive_header` and adds pass through.

        """

        func = self.parent.add_directive_header

        pass_through = ["__doc_disable_except__"]
        specific = getattr(self.parent, "pyautodoc_pass_to_directive", [])
        pass_through.extend(specific)

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            result = func(*args, **kwargs)
            for option in pass_through:
                self.pass_option_to_directive(option)

            return result

        self.parent.add_directive_header = wrapped

    def pass_option_to_directive(self, name: str):
        """Pass an autodoc option through to the generated directive.

        """

        if name in self.parent.options:
            source_name = self.parent.get_sourcename()
            value = self.parent.options[name]

            if isinstance("value", set):
                value = ", ".join(value)

            self.parent.add_line(f"   :{name}: {value}", source_name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sphinxcontrib', 'sphinxcontrib.autodoc_pydantic']

package_data = \
{'': ['*'], 'sphinxcontrib.autodoc_pydantic': ['css/*']}

install_requires = \
['Sphinx>=3.4', 'pydantic>=1.5']

extras_require = \
{'docs': ['sphinx-rtd-theme>=0.5.1,<0.6.0',
          'sphinx-tabs>=2,<3',
          'myst-parser>=0.13.7,<0.14.0']}

setup_kwargs = {
    'name': 'autodoc-pydantic',
    'version': '1.1.3',
    'description': 'Seamlessly integrate pydantic models in your Sphinx documentation.',
    'long_description': "![Autodoc Pydantic](docs/source/material/logo_black.svg)\n\n[![PyPI version](https://badge.fury.io/py/autodoc-pydantic.svg)](https://badge.fury.io/py/autodoc-pydantic)\n![Master](https://github.com/mansenfranzen/autodoc_pydantic/actions/workflows/tests.yml/badge.svg)\n![Python](https://img.shields.io/badge/python-3.6+-blue.svg)\n[![Codacy Badge](https://app.codacy.com/project/badge/Grade/30a083d784f245a98a0d5e6857708cc8)](https://www.codacy.com/gh/mansenfranzen/autodoc_pydantic/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=mansenfranzen/autodoc_pydantic&amp;utm_campaign=Badge_Grade)\n[![Codacy Badge](https://app.codacy.com/project/badge/Coverage/30a083d784f245a98a0d5e6857708cc8)](https://www.codacy.com/gh/mansenfranzen/autodoc_pydantic/dashboard?utm_source=github.com&utm_medium=referral&utm_content=mansenfranzen/autodoc_pydantic&utm_campaign=Badge_Coverage)\n[![Documentation Status](https://readthedocs.org/projects/autodoc-pydantic/badge/?version=latest)](https://autodoc-pydantic.readthedocs.io/en/latest/?badge=latest)\n\nYou love [pydantic](https://pydantic-docs.helpmanual.io/) :heart: and you want to document your models and configuration settings with [sphinx](https://www.sphinx-doc.org/en/master/)? \n\nPerfect, let's go. But wait, sphinx' [autodoc](https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html) does not integrate too well with pydantic models :confused:. \n\nDon't worry - just `pip install autodoc_pydantic` :relaxed:.\n\n## Features\n\n- :speech_balloon: provides default values, alias and constraints for model fields\n- :link: adds references between validators and corresponding fields\n- :page_with_curl: includes collapsable model json schema\n- :surfer: natively integrates with autodoc extension\n- :paperclip: defines explicit pydantic prefixes for models, settings, fields, validators and model config\n- :clipboard: shows summary section for model configuration and validators\n- :eyes: hides overloaded and redundant model class signature\n- :books: sorts fields, validators and model config within models by type\n- 🍀 Supports `pydantic >= 1.5.0` and `sphinx >= 3.4.0`\n\n### Comparison between autodoc sphinx and autodoc pydantic\n\n[![Comparison](docs/source/material/example_comparison_v1.0.0.gif)](https://autodoc-pydantic.readthedocs.io/en/latest/examples.html#default-configuration)\n\nTo see those features in action, jump over to the [example documentation](https://autodoc-pydantic.readthedocs.io/en/latest/examples.html#default-configuration) to compare\nthe appearance of standard sphinx autodoc with *autodoc_pydantic*.\n\n## Documentation\n\nFor more details, please visit the official [documentation](https://autodoc-pydantic.readthedocs.io/en/latest/):\n\n- [Installation](https://autodoc-pydantic.readthedocs.io/en/latest/installation.html)\n- [Configuration](https://autodoc-pydantic.readthedocs.io/en/latest/configuration.html)\n- [Usage](https://autodoc-pydantic.readthedocs.io/en/latest/usage.html)\n- [Examples](https://autodoc-pydantic.readthedocs.io/en/latest/examples.html)\n\n## Acknowledgements\n\nThanks to great open source projects [sphinx](https://www.sphinx-doc.org/en/master/), [pydantic](https://pydantic-docs.helpmanual.io/) and [poetry](https://python-poetry.org/) (and so many more) :heart: !\n",
    'author': 'mansenfranzen',
    'author_email': 'franz.woellert@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mansenfranzen/autodoc_pydantic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)
